/*
 * Copyright (C) 2004 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * Created by AJ Lewis <alewis@redhat.com> on June 17, 2004
 *
 */

/*  SCSI page80 and page83 code originally derived from and inspired by Doug
 *  Gilbert's scsi_inquiry.c, part of sg_utils package.
 *  See http://www.torque.net/sg
 *  Copyright (C) 2001 Compaq Computer Corporation
 *  Copyright (C) 1999 D. Gilbert
 */

/*  SCSI page80 and page83 derived from Steve Cameron's <smcameron@yahoo.com>
 *  implementation of scsi_serialno as part of his scsimap package.
 *  Original derivation from Doug Gilbert as stated above.
 *  Modifications by Gary Lerhaupt <gary_lerhaupt@dell.com>
 *  Copyright (C) 2002-2004 Dell Computer Corporation
 */

/* vpd check (do_scsi_vpdp_inquiry) coded by Dan Clark <dclark@aracnet.com>
 * Jan 06, 2004 - version 0.99
 */

/* most of this code was derived from scsi_unique_id in the devlabel package */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include "scsi-info.h"

#define DEVLABEL_PATH "/etc/sysconfig/devlabel"
#define LU_ASSOCIATION 0x00
#define PAGE80_SN_OFFSET 4

/* From scsi-unique-id ********/
#define SCSI_IOCTL_GET_IDLUN            0x5382
#define SCSI_IOCTL_GET_BUS_NUMBER       0x5386
#define INQUIRY_CMD     0x12
#define INQUIRY_CMDLEN  6
#define INQUIRY_REPLY_LEN 96
#define OFF (2 * sizeof(unsigned int))
#define PG83_IDL_OFFSET 4  /* Page 83 Identifier descriptor list offset */
#define PG83_IDL_ID_OFFSET 4  /* Page 83 Identifier descriptor list
			       * Identifier offset */

#define VPD_DEVICE_ID_PG           0x83
#define VPD_UNIT_SERIAL_NUM_PG     0x80
#define VPD_SUPPORTED_PAGES_PG     0x00

#ifndef SCSI_IOCTL_SEND_COMMAND
#define SCSI_IOCTL_SEND_COMMAND 1
#endif

typedef struct my_scsi_ioctl_command {
	unsigned int inlen;  /* _excluding_ scsi command length */
	unsigned int outlen;
	unsigned char data[1];  /* was 0 but that's not ISO C!! */
		/* on input, scsi command starts here then opt. data */
} My_Scsi_Ioctl_Command;

#define UNKNOWN_TYPE -1
#define SCSI_TYPE 0


static int
stat_device(int fd, unsigned int *rdev, int *devtype)
{
    struct stat statbuf;
    int res, to;

    *devtype = UNKNOWN_TYPE;
    res = fstat(fd, &statbuf);
    if (res != 0) {
        fprintf(stderr, "Cannot stat, %s\n", strerror(errno));
        return -1;
    }

    *rdev = statbuf.st_rdev;

    if (! S_ISBLK(statbuf.st_mode) ) {
        fprintf(stderr, "Not a block device!\n");
        return -1;
    }

    // Don't worry, being very careful not to write to a none-scsi file ...
    if (*devtype == UNKNOWN_TYPE) {
    	res = ioctl(fd, SCSI_IOCTL_GET_BUS_NUMBER, &to);
    	if (res < 0) {
            // perror("ioctl on scsi device, error");
            return -1;
    	}
	else
	*devtype = SCSI_TYPE;
    }

    return 0;
}

static int
generic_scsi_inquiry(int fd, unsigned char evpd, unsigned char page,
                     unsigned char *buf, unsigned char  buflen)
{
    unsigned char inqCmdBlk[INQUIRY_CMDLEN] =
        {INQUIRY_CMD, evpd, page, 0, INQUIRY_REPLY_LEN, 0};
    int rc, len;
    unsigned char *inqbuf;
    unsigned char *buffp;
    My_Scsi_Ioctl_Command * ishp;

    inqbuf = malloc(OFF + sizeof(inqCmdBlk) + 512);
    memset(inqbuf, 0, OFF + sizeof(inqCmdBlk) + 512);
    buffp = inqbuf + OFF;

    ishp = (My_Scsi_Ioctl_Command *) inqbuf;
    ishp->inlen = 0;
    ishp->outlen = INQUIRY_REPLY_LEN;
    memcpy(buffp, inqCmdBlk, INQUIRY_CMDLEN);
    rc = ioctl(fd, SCSI_IOCTL_SEND_COMMAND, inqbuf);
    if (rc != 0) {
        free(inqbuf);
        return rc;
    }
    len = INQUIRY_REPLY_LEN < buflen ?  INQUIRY_REPLY_LEN : buflen;
    memcpy(buf, buffp, len);
    free(inqbuf);
    return len;
}

/* Get the page83 device identification descriptor */
static int
do_scsi_83_inquiry(int fd, int rdev, struct scsi_info *info)
{
    int len, i, j, id_length, id_type, id_association;
    unsigned char buf[256];
    char tmp[2], **tmpid;
    int page_len;

    memset(buf, 0, 256);
    len = generic_scsi_inquiry(fd, 1, 0x83, buf, 255);
    if (len < 0) return len;

    page_len = buf[3]; /* FIXME: This seems wrong - the MSB of the
			* page length is in buf[2] according to the
			* SCSI spec - ajl(6/28/2004) */

    if (buf[1] != 0x83) {
        fprintf(stderr,"page83: No page83 information found.\n");
        return -1;
    } else {
        for (i=PG83_IDL_OFFSET; i < page_len + PG83_IDL_OFFSET; i++) {
            // Get info about the ID
            id_type = buf[i+1] & 0x0f;
            id_association = buf[i+1] & 0x30;
            id_length = buf[i+3];

	    switch(id_type) {
	    case 0:
		    tmpid = &info->page83_type0;
		    break;
	    case 1:
		    tmpid = &info->page83_type1;
		    break;

	    case 2:
		    tmpid = &info->page83_type2;
		    break;

	    case 3:
		    tmpid = &info->page83_type3;
		    break;
	    default:
		    /* Set tmpid so we skip allocating on it */
		    tmpid = (char **) 1;
		    break;
	    }
            // Print out the Identifier
            i = i + PG83_IDL_ID_OFFSET;
            if (id_association == LU_ASSOCIATION && !*tmpid) {
                if (i+id_length > len) {
                    fprintf(stderr, "Device returned an invalid VPD while reading page83\n");
                    break;
                }
		/* Each byte is represented by a 2 hex digits */
                if(!(*tmpid = malloc(sizeof(char) * ((id_length * 2) + 1)))) {
                    fprintf(stderr, "Unable to allocate memory for page83\n");
                    return -1;
                }
                if(!memset(*tmpid, 0, sizeof(**tmpid))) {
                    fprintf(stderr, "Unable to zero memory for page80\n");
                    return -1;
                }
                for (j=0; j < id_length; j++) {
                    sprintf(tmp, "%02x", buf[i+j]);
                    strcat(*tmpid, tmp);
                }
            } else {
                j = id_length;
            }

            // Increment i but subtract 1 to compensate for j loop
            i=i+j-1;
        }

        return 0;
    }
}

/* Grab the Product Serial Number out of page80 */
static int
do_scsi_80_inquiry(int fd, int rdev, struct scsi_info *info)
{
    int len, i, id_length;
    unsigned char buf[256];
    char tmp[2];

    memset(buf, 0, 256);
    len = generic_scsi_inquiry(fd, 1, 0x80, buf, 255);
    if (len < 0) return len;

    if (buf[1] != 0x80) {
        fprintf(stderr,"page80: No page80 information found.\n");
        return -1;
    } else {
        // Print out the Serial #
        id_length = buf[3];
        /* Each byte is represented by a 2 hex digits */
        if(!(info->page80 = malloc(sizeof(char) * ((id_length * 2) + 1)))) {
            fprintf(stderr, "Unable to allocate memory for page80\n");
            return -1;
        }
        if(!memset(info->page80, 0, sizeof(*info->page80))) {
            fprintf(stderr, "Unable to zero memory for page80\n");
            return -1;
        }
        if (PAGE80_SN_OFFSET + id_length > len) {
            fprintf(stderr, "Buffer overrun while reading scsi page80");
            return -1;
        }
        for (i = PAGE80_SN_OFFSET; i < id_length + PAGE80_SN_OFFSET; i++) {
            sprintf(tmp, "%02x", buf[i]);
            strcat(info->page80, tmp);
        }
        return 0;
    }
}

/*
 * inquire for the vital product data pages
 * determine if 0x80 and 0x83 inquiries are supported or not
 * then make inquiry of pages that are supported by device/driver
 * 0x00 supported vital product data pages
 * 0x80 unit serial number page
 * 0x83 device identification page
 */
static int
do_scsi_vpdp_inquiry(int fd, int rdev, struct scsi_info *info)
{
    int len, i;
    unsigned char buf[256];
    ushort vpdp_cnt;        /* count of vpdp pages */

    memset(buf, 0, 256);
    len = generic_scsi_inquiry(fd, 1, VPD_SUPPORTED_PAGES_PG, buf, 255);
    if (len < 0) return len;
    vpdp_cnt = (ushort)buf[3];
    for(i = 0; i < vpdp_cnt; i++) {
        ushort vpdp_type;
        vpdp_type = buf[4 + i];

        switch(vpdp_type) {
        case VPD_UNIT_SERIAL_NUM_PG:
            // Get SCSI page80 unite serial number page
            do_scsi_80_inquiry(fd, rdev, info);
            break;
        case VPD_DEVICE_ID_PG:
            // Get SCSI page83 device identification page
            do_scsi_83_inquiry(fd, rdev, info);
            break;
        default:
            break;
        }
    }

    return 0;
}


int
get_scsi_info(struct scsi_info *info)
{
    int fd;
    char realdev[PATH_MAX + 1];
    u_long arg[2];
    unsigned int rdev;
    int devtype;

    /* get the (host,channel,id,lun) quadruple for this device, if possible */
    if ((fd = open(info->path, O_RDONLY)) >= 0) {

	/* try to resolve any symlinks out of the name */
	if (realpath(info->path, realdev)) {
	    realdev[sizeof (realdev) - 1] = '\0';
            free(info->path);
	    info->path = strdup(realdev);
	}

	memset(&arg, 0, sizeof (arg));
	/* try to get the host,channel,id,port quadruple */
	if (ioctl(fd, SCSI_IOCTL_GET_IDLUN, &arg) == 0) {
	    info->id = arg[0] & 0xff;
	    info->lun = (arg[0] >> 8) & 0xff;
	    info->channel = (arg[0] >> 16) & 0xff;
	    info->host = (arg[0] >> 24) & 0xff;
#ifdef SCSI_IOCTL_GET_BUS_NUMBER
	    /* 2.4 kernels give us all the info we need with
	     * that ioctl.  2.2 kernels have another ioctl
	     * to get the host number, and set the host
	     * above to something useless to us.  If this
	     * 2nd ioctl fails, assume it's a 2.4 kernel,
	     * and use the host number we already have.
	     */
	    memset(&arg, 0, sizeof (arg));
	    if (ioctl(fd, SCSI_IOCTL_GET_BUS_NUMBER, &arg) == 0) {
		info->host = arg[0];
	    }
#endif
	} else {
	    fprintf(stderr,
		    "iscsi-device: ioctl SCSI_IOCTL_GET_IDLUN %s failed, %s\n",
		    info->path, strerror(errno));
	    exit(1);
	}
        if (stat_device(fd, &rdev, &devtype) != 0) {
            fprintf(stderr, "%s: Unknown devtype on %s\n", "foo", info->path);
            exit(1);
        }
        do_scsi_vpdp_inquiry(fd, rdev, info);
        close(fd);
    } else {
	fprintf(stderr, "iscsi-device: couldn't open %s, %s\n", info->path,
		strerror(errno));
	exit(1);
    }
    return 1;
}

static int
dl_multipath(char *params)
{
    if(strstr(params, "multipath"))
        return 1;
    return 0;
}

static char *
find_mp_link(char *link, char *path)
{

    char mppath[PATH_MAX], tmppath[PATH_MAX], rp[PATH_MAX];
    char number[4]; /* If we have more than 999 paths... */
    int i = 0;
    strcpy(mppath, link);
    strcat(mppath, "_multipath");

    /* FIXME: 999 is abitrary, but we should almost never get there
     * since if the mppath$i doesn't exist, we return NULL - this also
     * means that if for some reason one symlink in the middle is
     * missing, we never find the rest */
    for (i = 0; i < 999; i++) {
	sprintf(number, "%d", i);
	strcpy(tmppath, mppath);
	strcat(tmppath, number);
	if(!realpath(tmppath, rp))
	    return NULL;
	if(!strcmp(rp, path))
	    return strdup(tmppath);
    }
    return NULL;

}

char *
get_link_name(char *params, char *link, char *expected_path) {

    char *mylink;
    char path[PATH_MAX+1];

    if(dl_multipath(params)) {
	mylink = find_mp_link(link, expected_path);
	if(!mylink) {
	    fprintf(stderr, "Warning: devlabel multipath symlinks not setup\n");
	    return mylink;
	}
    }
    else
	mylink = strdup(link);

    if(mylink) {
	if(!realpath(mylink, path))
	    fprintf(stderr, "Warning: devlabel symlink (%s)"
		    " not setup\n", mylink);
	else {
	    if(strcmp(path, expected_path))
		fprintf(stderr, "Warning: devlabel linked device"
			" name (%s) does not match discovered"
			" device name (%s)\n", path, expected_path);
	}
    }

    return mylink;
}

char *
get_devlabel_link(struct scsi_info *info)
{

    FILE *fp;
    char *rdbufp, rdbuf[1024], *tmp, *link = NULL;
    /* FIXME: PATH_MAX doesn't necessarily make sense for uuid and params */
    char tmplink[PATH_MAX], path[PATH_MAX+1], uuid[PATH_MAX], params[PATH_MAX];
    char type[16];

    if(!info->page83_type3 && !info->page83_type2 && !info->page83_type1 &&
       !info->page80 && !info->page83_type0) {
        fprintf(stderr, "No uuid detected for device %s", info->path);
        return NULL;
    }

    fp = fopen(DEVLABEL_PATH, "r");
    if(!fp) {
        perror("Cannot open " DEVLABEL_PATH);
        return NULL;
    }

    do {
        rdbufp = fgets(rdbuf, sizeof (rdbuf), fp);
	if (rdbufp && (*rdbufp != '#')) {
	    while (isspace(*rdbufp)) {
		rdbufp += 1;
	    }
            if(rdbufp[0] == '\n' || rdbufp[0] == '\0')
                continue;
	    /* link name */
            tmp = strchr(rdbufp, ' ');
	    if(!tmp)
		continue;
            tmp[0] = '\0';
            strcpy(tmplink, rdbufp);
            rdbufp = tmp+1;
	    /* Device name */
            tmp = strchr(rdbufp, ' ');
	    if(!tmp)
		continue;
            tmp[0] = '\0';
            strcpy(path, rdbufp);
            rdbufp = tmp+1;
	    /* Grab the type */
            tmp = strchr(rdbufp, ':');
	    if(!tmp)
		continue;
            tmp[0] = '\0';
            strcpy(type, rdbufp);
            rdbufp = tmp+1;
	    /* Check for multipath */
	    tmp = strchr(rdbufp, ' ');
	    if(tmp) {
		tmp[0] = '\0';
	    }
	    /* Read in the uuid */
	    strcpy(uuid, rdbufp);
	    if(tmp) {
		rdbufp = tmp+1;
		/*read in the last word */
		strcpy(params, rdbufp);
	    }
            if(!strcmp(type, "S83.3") && info->page83_type3) {
		if(!strncmp(uuid, info->page83_type3,
			    strlen(info->page83_type3))) {
		    link = get_link_name(params, tmplink, info->path);
		    break;
                }
                else
                    continue;
            }
	    if(!strcmp(type, "S83.2") && info->page83_type2) {
                if(!strncmp(uuid, info->page83_type2, strlen(info->page83_type2))) {
		    link = get_link_name(params, tmplink, info->path);
		    break;
                }
                else
                    continue;
            }
	    if(!strcmp(type, "S83.1") && info->page83_type1) {
                if(!strncmp(uuid, info->page83_type1, strlen(info->page83_type1))) {
		    link = get_link_name(params, tmplink, info->path);
                    break;
                }
                else
                    continue;
            }
            if(!strncmp(type, "S80", 3) && info->page80) {
                if(!strncmp(uuid, info->page80, strlen(info->page80))) {
		    link = get_link_name(params, tmplink, info->path);
                    break;
                }
                else continue;
            }
	    if(!strcmp(type, "S83.0") && info->page83_type0) {
                if(!strncmp(uuid, info->page83_type0, strlen(info->page83_type0))) {
		    link = get_link_name(params, tmplink, info->path);
                    break;
                }
                else
                    continue;
            }
        }
    } while(rdbufp);
    fclose(fp);
    return link;
}
